{% extends "layout.html" %}
{% block body %}
<h3>
    Welcome to GNS3.
</h3>
<ul>
    <li><a href="https://gns3.com">Website</a></li>
    <li><a href="http://api.gns3.net">API documentation</a></li>
</ul>
<p>If you are looking for uploading the IOU. You can since 1.4 upload them directly from the client see: <a href="https://gns3.com/support/docs/how-to-configure-non-native-io-3">this documenation</a>.</p>
{% endblock %}
