The upload form is deprecated since GNS3 1.4.<br>
GNS3 will upload the images for you when you select it in the interface.
<br><br>
For the iourc file you just need to configure the file in the IOU section of the GUI.
<br><br>
More informations on the documentation section of <a href="http://www.gns3.com">GNS3.com</a>
