<html>
<head> 
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8"/> 
{% block head %}{% endblock %}
<title>GNS3 Server</title>
</head>
<body>
<div>
    <a href="/">Home</a>
    |
    <a href="/controller">Controller status</a>
    |
    <a href="/compute">Compute status</a>
</div>
{% block body %}{% endblock %}
</body>
<small>
    Powered by GNS3 {{gns3_version}}
</small>
</html>
